import typing as t
from typing import Any
from typing import Optional

import gcsfs
from google.oauth2.credentials import Credentials


class GCSFileSystem(gcsfs.GCSFileSystem):  # type: ignore [misc]
    """GCSFileSystem is a wrapper around gcsfs.GCSFileSystem."""

    def __init__(
        self, token: Optional[dict[str, str] | str | Credentials] = None, **kwargs: Any
    ) -> None:
        """Initialize GCSFileSystem."""
        super().__init__(token=token, **kwargs)

    def isdir(self, path: str) -> bool:
        """Check if path is a directory."""
        info = super(gcsfs.GCSFileSystem, self).info(path)
        return t.cast(bool, info["type"] == "directory")



import typing as t
from typing import Any
from typing import Optional

import requests

from .auth import AuthClient


class GuardianClient:
    """Client for interacting with the Maskinporten Guardian."""

    @staticmethod
    def call_api(
        api_endpoint_url: str,
        maskinporten_client_id: str,
        scopes: str,
        guardian_endpoint_url: str = "http://maskinporten-guardian.dapla.svc.cluster.local/maskinporten/access-token",
        keycloak_token: Optional[str] = None,
    ) -> Any:
        """Call an external API using Maskinporten Guardian.

        Args:
            api_endpoint_url: URL to the target API
            maskinporten_client_id: the Maskinporten client id
            scopes: the Maskinporten scopes
            guardian_endpoint_url: URL to the Maskinporten Guardian
            keycloak_token: the user's personal Keycloak token. Automatic fetch attempt will be made if left empty.

        Raises:
            RuntimeError: If the API call fails

        Returns:
            The endpoint json response
        """
        if keycloak_token is None:
            keycloak_token = AuthClient.fetch_personal_token()
        body = {"maskinportenClientId": maskinporten_client_id, "scopes": scopes}
        maskinporten_token = GuardianClient.get_guardian_token(
            guardian_endpoint_url, keycloak_token, body=body
        )
        api_response = requests.get(
            api_endpoint_url,
            headers={
                "Authorization": f"Bearer {maskinporten_token}",
                "Accept": "application/json",
            },
        )
        if api_response.status_code == 200:
            return api_response.json()
        else:
            raise RuntimeError(
                f"Error calling target API ({api_endpoint_url}): Status code <{api_response.status_code}"
                f" - {api_response.text or api_response.reason}>"
            )

    @staticmethod
    def get_guardian_token(
        guardian_endpoint: str, keycloak_token: str, body: dict[str, str]
    ) -> str:
        """Retrieve access token from Maskinporten Guardian.

        Args:
            guardian_endpoint: URL to the maskinporten guardian
            keycloak_token: the user's Keycloak token
            body: maskinporten request body

        Raises:
            RuntimeError: If the Guardian token request fails

        Returns:
            The maskinporten access token
        """
        guardian_response = requests.post(
            guardian_endpoint,
            headers={
                "Authorization": f"Bearer {keycloak_token}",
                "Content-type": "application/json",
            },
            json=body,
        )
        if guardian_response.status_code == 200:
            return t.cast(str, guardian_response.json()["accessToken"])
        else:
            raise RuntimeError(
                f"Error getting guardian token: <{guardian_response.status_code}: "
                f"{guardian_response.text or guardian_response.reason}>"
            )



from pathlib import Path
from typing import Optional


def repo_root_dir(directory: Optional[Path | str] = None) -> Path:
    """Find the root directory of a git repo, searching upwards from a given path.

    Args:
        directory: The path to search from, defaults to the current working directory.
            The directory can be of type string or of type pathlib.Path.

    Returns:
        Path to the git repo's root directory.

    Raises:
        RuntimeError: If no .git directory is found when searching upwards.

    Example:
    --------
    >>> import dapla as dp
    >>> import tomli
    >>>
    >>> config_file = dp.repo_root_dir() / "pyproject.toml"
    >>> with open(config_file, mode="rb") as fp:
    >>>     config = tomli.load(fp)
    """
    if directory is None:
        directory = Path.cwd()

    if isinstance(directory, str):
        directory = Path(directory)

    while directory / ".git" not in directory.iterdir():
        if directory == Path("/"):
            raise RuntimeError(f"The directory {directory} is not in a git repo.")
        else:
            directory = directory.parent
    return directory



import typing as t
from io import TextIOWrapper
from typing import Any

import google
import pandas as pd
from fsspec.spec import AbstractBufferedFile
from google.cloud import storage

from .auth import AuthClient
from .gcs import GCSFileSystem

GS_URI_PREFIX = "gs://"


class FileClient:
    """Client for working with buckets and files on Google Cloud Storage.

    This class should not be instantiated, only the static methods should be used.
    """

    @staticmethod
    def _ensure_gcs_uri_prefix(gcs_path: str) -> str:
        """Ensure that a GCS uri has the 'gs://' prefix.

        Some operations require GCS uris, but we don't want the user to bother with knowing when to use the prefix,
        so we ensure its presence automatically where it is needed.
        """
        if not gcs_path.startswith(GS_URI_PREFIX):
            gcs_path = f"{GS_URI_PREFIX}{gcs_path}"
        return gcs_path

    @staticmethod
    def _remove_gcs_uri_prefix(gcs_path: str) -> str:
        """Remove the 'gs://' prefix from a GCS URI."""
        if gcs_path.startswith(GS_URI_PREFIX):
            gcs_path = gcs_path[len(GS_URI_PREFIX) :]
        return gcs_path

    @staticmethod
    def get_gcs_file_system(**kwargs: Any) -> GCSFileSystem:
        """Return a pythonic file-system for Google Cloud Storage - initialized with a personal Google Identity token.

        Args:
            kwargs: Additional arguments to pass to the underlying GCSFileSystem.

        Returns:
            A GCSFileSystem instance.

        See https://gcsfs.readthedocs.io/en/latest for advanced usage
        """
        return GCSFileSystem(token=AuthClient.fetch_google_credentials(), **kwargs)

    @staticmethod
    def ls(gcs_path: str, detail: bool = False, **kwargs: Any) -> Any:
        """List the contents of a GCS bucket path.

        Args:
            gcs_path: The GCS path to a directory.
            detail: Whether to return detailed information about the files.
            kwargs: Additional arguments to pass to the underlying 'ls()' method.

        Returns:
            List of strings if detail is False, or list of directory information dicts if detail is True.
        """
        return FileClient.get_gcs_file_system().ls(gcs_path, detail=detail, **kwargs)

    @staticmethod
    def get_versions(bucket_name: str, file_path: str) -> Any:
        """Get all versions of a file in a bucket.

        Args:
            bucket_name: Bucket name where the file is located.
            file_path: Path to the file.

        Returns:
            List of versions of the file.
        """
        storage_client = storage.Client(
            credentials=AuthClient.fetch_google_credentials()
        )
        try:
            bucket = storage_client.get_bucket(bucket_name)

            if bucket.versioning_enabled:
                return list(bucket.list_blobs(prefix=file_path, versions=True))
            else:
                return list(bucket.list_blobs(prefix=file_path, soft_deleted=True))
        except google.api_core.exceptions.NotFound:
            print(f'Bucket "{bucket_name}" does not exist')
            return []

    @staticmethod
    def restore_version(
        source_bucket_name: str,
        source_file_name: str,
        source_generation_id: str,
        **kwargs: Any,
    ) -> Any:
        """Restores soft deleted/object versionated version of file to the live version.

        Args:
            source_bucket_name: source bucket name where the file is located.
            source_file_name: non-current file name.
            source_generation_id: generation_id of the non-current.
            kwargs: Additional arguments to pass to the underlying 'copy_blob()' method.

        Returns:
            A new blob with new generation id.
        """
        storage_client = storage.Client(
            credentials=AuthClient.fetch_google_credentials()
        )

        try:
            source_bucket = storage_client.get_bucket(source_bucket_name)
            source_file = source_bucket.blob(source_file_name)

            if source_bucket.versioning_enabled:
                try:
                    return source_bucket.copy_blob(
                        blob=source_file,
                        destination_bucket=source_bucket,
                        source_generation=source_generation_id,
                        **kwargs,
                    )
                except google.api_core.exceptions.NotFound:
                    print(
                        f'No such object "{source_file_name}" exist with generationnumber "{source_generation_id}".'
                    )
                    return []
            else:
                try:
                    return source_bucket.restore_blob(
                        blob_name=source_file_name,
                        generation=source_generation_id,
                        **kwargs,
                    )
                except google.api_core.exceptions.NotFound:
                    print(
                        f'No such object "{source_file_name}" exist with generationnumber "{source_generation_id}"'
                    )
                    return []

        except google.api_core.exceptions.NotFound:
            print(f'Bucket "{source_bucket_name}" does not exist')

    @staticmethod
    def cat(gcs_path: str) -> str:
        """Get string content of a file from GCS.

        Args:
            gcs_path: The GCS path to a file.

        Returns:
            utf-8 decoded string content of the given file
        """
        return t.cast(
            str, FileClient.get_gcs_file_system().cat(gcs_path).decode("utf-8")
        )

    @staticmethod
    def gcs_open(
        gcs_path: str, mode: str = "r"
    ) -> TextIOWrapper | AbstractBufferedFile:
        """Open a file in GCS, works like regular python open().

        Args:
            gcs_path: The GCS path to a file.
            mode: File open mode. Defaults to 'r'

        Returns:
            A file-like object.
        """
        return FileClient.get_gcs_file_system().open(
            FileClient._ensure_gcs_uri_prefix(gcs_path), mode
        )

    @staticmethod
    def load_csv_to_pandas(gcs_path: str, **kwargs: Any) -> pd.DataFrame:
        """Reads a CSV file from Google Cloud Storage into a Pandas DataFrame.

        Args:
            gcs_path: The GCS path to a .csv file.
            **kwargs: Additional arguments to pass to the underlying Pandas read_csv().

        Returns:
            A Pandas DataFrame.
        """
        return t.cast(
            pd.DataFrame,
            pd.read_csv(
                FileClient._ensure_gcs_uri_prefix(gcs_path),
                storage_options={"token": AuthClient.fetch_google_credentials()},
                **kwargs,
            ),
        )

    @staticmethod
    def load_json_to_pandas(gcs_path: str, **kwargs: Any) -> pd.DataFrame:
        """Reads a JSON file from Google Cloud Storage into a Pandas DataFrame.

        Args:
            gcs_path: The GCS path to a .json file.
            **kwargs: Additional arguments to pass to the underlying Pandas read_json().

        Returns:
            A Pandas DataFrame.
        """
        return t.cast(
            pd.DataFrame,
            pd.read_json(
                FileClient._ensure_gcs_uri_prefix(gcs_path),
                storage_options={"token": AuthClient.fetch_google_credentials()},
                **kwargs,
            ),
        )

    @staticmethod
    def load_xml_to_pandas(gcs_path: str, **kwargs: Any) -> pd.DataFrame:
        """Reads an XML file from Google Cloud Storage into a Pandas DataFrame.

        Args:
            gcs_path: The GCS path to a .xml file.
            **kwargs: Additional arguments to pass to the underlying Pandas read_xml().

        Returns:
            A Pandas DataFrame.
        """
        return pd.read_xml(
            FileClient._ensure_gcs_uri_prefix(gcs_path),
            storage_options={"token": AuthClient.fetch_google_credentials()},
            **kwargs,
        )

    @staticmethod
    def save_pandas_to_csv(df: pd.DataFrame, gcs_path: str, **kwargs: Any) -> None:
        """Write the contents of a Pandas DataFrame to a CSV file in a bucket.

        Args:
            df: The Pandas DataFrame to save to file.
            gcs_path: The GCS path to the destination .csv file.
            **kwargs: Additional arguments to pass to the underlying Pandas to_csv().
        """
        df.to_csv(
            FileClient._ensure_gcs_uri_prefix(gcs_path),
            storage_options={"token": AuthClient.fetch_google_credentials()},
            **kwargs,
        )

    @staticmethod
    def save_pandas_to_json(df: pd.DataFrame, gcs_path: str, **kwargs: Any) -> None:
        """Write the contents of a Pandas DataFrame to a JSON file in a bucket.

        Args:
            df: The Pandas DataFrame to save to file.
            gcs_path: The GCS path to the destination .json file.
            **kwargs: Additional arguments to pass to the underlying Pandas to_json().
        """
        df.to_json(
            FileClient._ensure_gcs_uri_prefix(gcs_path),
            **kwargs,
        )

    @staticmethod
    def save_pandas_to_xml(df: pd.DataFrame, gcs_path: str, **kwargs: Any) -> None:
        """Write the contents of a Pandas DataFrame to an XML file in a bucket.

        Args:
            df: The Pandas DataFrame to save to file.
            gcs_path: The GCS path to the destination .xml file.
            **kwargs: Additional arguments to pass to the underlying Pandas to_xml().
        """
        df.to_xml(
            FileClient._ensure_gcs_uri_prefix(gcs_path),
            storage_options={"token": AuthClient.fetch_google_credentials()},
            **kwargs,
        )
